#include "StdAfx.h"
#include "ExposeAnimator.h"
#include "math.h"
//using namespace
using namespace System::Collections::Generic;

ExposeAnimator::ExposeAnimator(HWND target, voiddelegate^ animEndCallback)
{
	this->m_hTarget = target;
	this->m_fEndCallback = animEndCallback;
	m_nAnimState = EA_NONE;
	m_nLastWindowIndex = -1;
}



ExposeAnimator::~ExposeAnimator(void)
{
	Reset();
}

void ExposeAnimator::OnRender()
{
	

	if(m_nAnimState != EA_NONE)
	{
		LARGE_INTEGER current;
		QueryPerformanceCounter(&current);
		Window * w;
		RECT rSrc;
		DWM_THUMBNAIL_PROPERTIES props;
		props.dwFlags = DWM_TNP_RECTDESTINATION;
		if(current.QuadPart > m_nAnimEnd)
		{
			if(m_nAnimState & EA_STARTING)
			{
				for(SIZE_T i = 0; i< m_lWindows.size();i++)
				{
					w = m_lWindows.at(i);
					rSrc = w->WindowSourceRECT;
					props.rcDestination.bottom = rSrc.bottom+w->targetOffsetBottom;
					props.rcDestination.top = rSrc.top+w->targetOffsetTop;
					props.rcDestination.left = rSrc.left+w->targetOffsetLeft;
					props.rcDestination.right = rSrc.right+w->targetOffsetRight;
					

					DwmUpdateThumbnailProperties(w->ThumbID,&props);
				}
				m_nAnimState-=EA_STARTING;
			}
			else if(m_nAnimState & EA_ENDING)
			{
				for(SIZE_T i = 0; i< m_lWindows.size();i++)
				{
					w = m_lWindows.at(i);
					props.rcDestination = w->WindowSourceRECT;
					DwmUpdateThumbnailProperties(w->ThumbID,&props);
				}
				this->Reset();
				this->m_fEndCallback->Invoke();
				m_nAnimState-=EA_ENDING;
			}
		
		}
		else
		{
			double offset =  (( double)(current.QuadPart - m_nAnimStart)) / ((double) m_nDuration);
			if(m_nAnimState & EA_STARTING)
			{
				for(SIZE_T i = 0; i<m_lWindows.size();i++)
				{
					w = m_lWindows.at(i);
					rSrc = w->WindowSourceRECT;

					props.rcDestination.bottom = rSrc.bottom+ (int)(w->targetOffsetBottom * offset);
					props.rcDestination.top = rSrc.top+(int)(w->targetOffsetTop * offset);
					props.rcDestination.left = rSrc.left+(int)(w->targetOffsetLeft * offset);
					props.rcDestination.right = rSrc.right+(int)(w->targetOffsetRight * offset);
					

					DwmUpdateThumbnailProperties(w->ThumbID,&props);
				}
			}
			else if(m_nAnimState & EA_ENDING)
			{
				for(SIZE_T i = 0; i<m_lWindows.size();i++)
				{
					w = m_lWindows.at(i);
					rSrc = w->WindowSourceRECT;

					props.rcDestination.bottom = rSrc.bottom+w->targetOffsetBottom-(int)(w->targetOffsetBottom * offset);
					props.rcDestination.top = rSrc.top+w->targetOffsetTop-(int)(w->targetOffsetTop * offset);
					props.rcDestination.left = rSrc.left+w->targetOffsetLeft-(int)(w->targetOffsetLeft * offset);
					props.rcDestination.right = rSrc.right+w->targetOffsetRight-(int)(w->targetOffsetRight * offset);
					

					DwmUpdateThumbnailProperties(w->ThumbID,&props);
				}
			}
		}
	}
}
void ExposeAnimator::Reset()
{
	for(SIZE_T i =0; i < m_lWindows.size();i++)
	{
		DwmUnregisterThumbnail(m_lWindows.at(i)->ThumbID);
		delete m_lWindows.at(i);
	}
	this->m_lWindows.clear();
}
void ExposeAnimator::AbortExpose()
{

	this->InitAnimation(EA_ENDING);
	//this->Reset();
	//this->m_fEndCallback->Invoke();
}

void ExposeAnimator::StartExpose()
{
	this->GetVisibleWindows();
	this->ComputeTargets();
	this->InitAnimation(EA_STARTING);
}

void ExposeAnimator::InitAnimation(int animState)
{
	LARGE_INTEGER start;
	QueryPerformanceCounter(&start);
	m_nAnimStart = start.QuadPart;
	LARGE_INTEGER freq;
	QueryPerformanceFrequency(&freq);
	m_nAnimEnd = m_nAnimStart + (freq.QuadPart / 2);
	m_nDuration = m_nAnimEnd - m_nAnimStart;
	m_nAnimState |= animState;
}
void ExposeAnimator::InitFadings()
{
	LARGE_INTEGER start;
	QueryPerformanceCounter(&start);
	m_nFadingStart = start.QuadPart;
	LARGE_INTEGER freq;
	QueryPerformanceFrequency(&freq);
	m_nFadingEnd = m_nFadingStart + (freq.QuadPart / 2);
	m_nFadingDuration = m_nFadingEnd - m_nFadingStart;
	m_nAnimState |= EA_FADINGS;
}

void ExposeAnimator::GetVisibleWindows()
{
	HWND current = GetWindow(m_hTarget,GW_HWNDLAST);
	while(current != NULL)
	{
		
		if(current != m_hTarget && (TASKSTYLE == (TASKSTYLE & GetWindowLong(current,GWL_STYLE))))
		{
			Window * w = new Window() ;
			w->WindowHandle = current;
			GetWindowRect(current, &(w->WindowSourceRECT));
			//m_lWindows.AddTail(w);
			DwmRegisterThumbnail(m_hTarget,current, &(w->ThumbID));
			DWM_THUMBNAIL_PROPERTIES props;
			props.fVisible = TRUE;
			props.opacity = 180;
			
			props.rcDestination = w->WindowSourceRECT;
			props.dwFlags = DWM_TNP_VISIBLE | DWM_TNP_OPACITY | DWM_TNP_RECTDESTINATION;

			DwmUpdateThumbnailProperties(w->ThumbID,&props);
			w->opacity = 180;
			w->baseOpacity=180;
			m_lWindows.push_back(w);
		}
		current = GetWindow(current,GW_HWNDPREV);
	}
}

void ExposeAnimator::ComputeTargets()
{
	
	m_nWindowsCount= (long)m_lWindows.size();
	RECT wndRect;
	GetWindowRect(m_hTarget,&wndRect);
	m_nDepth = (long)ceil(pow(m_nWindowsCount,0.5));
	m_nCellWidth = (wndRect.right-wndRect.left) / m_nDepth;
	m_nCellHeight = (wndRect.bottom-wndRect.top) / m_nDepth;
	long hMargin = m_nCellWidth / 10;
	long vMargin = m_nCellHeight / 10;
	int col=0,row=0,rowpart=0;


	for(;row<m_nDepth;row++)
	{
		rowpart = row*m_nDepth;
		for(col=0;col<m_nDepth;col++)
		{
			if((rowpart+col) >= m_nWindowsCount)
				break;
			else
			{
				Window * w= m_lWindows.at(rowpart+col);
				w->targetOffsetLeft = col * m_nCellWidth + hMargin - w->WindowSourceRECT.left;
				w->targetOffsetRight = (col+1) * m_nCellWidth - hMargin - w->WindowSourceRECT.right;
				w->targetOffsetTop = row*m_nCellHeight + vMargin - w->WindowSourceRECT.top;
				w->targetOffsetBottom = (row+1)*m_nCellHeight -vMargin - w->WindowSourceRECT.bottom;
			}
		}
	}
}
long ExposeAnimator::hitTest(double x, double y)
{
	long col = (long)x / m_nCellWidth;
	long row = (long)y / m_nCellHeight;
	return row * m_nDepth + col;
}
void ExposeAnimator::OnMouseMove(double x, double y)
{
	if(m_nAnimState & EA_STARTING || m_nAnimState & EA_ENDING)
		return;
	DWM_THUMBNAIL_PROPERTIES props;
	props.dwFlags = DWM_TNP_OPACITY;
	long index = hitTest(x,y);
	if(m_nLastWindowIndex != index)
	{
		if(m_nLastWindowIndex<m_nWindowsCount && m_nLastWindowIndex>-1)
		{
			m_nFadingOut = m_nLastWindowIndex;
			Window *w;
			w= m_lWindows.at(m_nFadingOut);
			w->baseOpacity = w->opacity;

			//TODO :  replace by Animation
			props.opacity = 180;
			DwmUpdateThumbnailProperties(w->ThumbID,&props);
		}
		else
		{
			m_nFadingOut = -1;
		}
		if(index<m_nWindowsCount)
		{
			m_nFadingIn = index;
			Window *w;
			w= m_lWindows.at(m_nFadingIn);
			w->baseOpacity = w->opacity;
			//TODO :  replace by Animation
			props.opacity = 255;
			DwmUpdateThumbnailProperties(w->ThumbID,&props);
		}
		else
		{
			m_nFadingIn = -1;
		}
		m_nLastWindowIndex = index;
		InitFadings();
	}
}
void ExposeAnimator::OnMouseClick(double x, double y)
{
	if(m_nAnimState & EA_STARTING || m_nAnimState & EA_ENDING)
		return;
	long index = hitTest(x,y);
	if(index < m_nWindowsCount)
	{
		Window * w = m_lWindows.at(index);
		SetForegroundWindow(w->WindowHandle);
		WINDOWINFO infos;
		GetWindowInfo(w->WindowHandle,&infos);
		if(infos.dwWindowStatus == 0)
			ShowWindow(w->WindowHandle,SW_RESTORE);
		//	ShowWindow(w->WindowHandle,SW_RESTORE);
		DwmUnregisterThumbnail(w->ThumbID);
		
		DwmRegisterThumbnail(m_hTarget, w->WindowHandle,&w->ThumbID);
		DWM_THUMBNAIL_PROPERTIES props;
		props.dwFlags = DWM_TNP_VISIBLE;
		props.dwFlags = DWM_TNP_RECTDESTINATION;
		props.fVisible = true;
		RECT r = w->WindowSourceRECT;
		r.bottom+=w->targetOffsetBottom;
		r.left += w->targetOffsetLeft;
		r.right += w->targetOffsetRight;
		r.top += w->targetOffsetTop;
		props.rcDestination = r;
		DwmUpdateThumbnailProperties(w->ThumbID,&props);
		InitAnimation(EA_ENDING);
	}
}